"""
file: header.py
brief: Defines how to handle PLD headers
author: S. V. Paulauskas
date: January 26, 2019
"""
import struct

import constants.data as data


class PldHeader:
    def __init__(self):
        self.run_number = self.max_spill_size = self.run_time = 0
        self.format = self.facility = self.start_date = self.end_date = self.title = ''

    def read_header(self, stream):
        """ Reads the header from a UTK PLD file. """
        return {
            'run_number': struct.unpack('I', stream.read(data.WORD))[0],
            'max_spill_size': struct.unpack('I', stream.read(data.WORD))[0],
            'run_time': struct.unpack('I', stream.read(data.WORD))[0],
            'format': stream.read(data.WORD * 4).decode(),
            'facility': stream.read(data.WORD * 4).decode(),
            'start_date': stream.read(data.WORD * 6).decode(),
            'end_date': stream.read(data.WORD * 6).decode(),
            'title': stream.read(struct.unpack('I', stream.read(data.WORD))[0]).decode()
        }



"""
file: list_mode_data_decoder.py
brief: Decodes binary data produced by an XIA Pixie16 module
author: S. V. Paulauskas
date: January 17, 2019
"""
from functools import partial
import struct
import threading

import dolosse.constants.data as data


def decode_word_zero(word, mask):
    """

    :param word: The word that we're going to decode
    :return: A dictionary containing the decoded information.
    """
    return {
        'channel': (word & mask.channel()[0]) >> mask.channel()[1],
        'slot': (word & mask.slot()[0]) >> mask.slot()[1],
        'crate': (word & mask.crate()[0]) >> mask.crate()[1],
        'header_length': (word & mask.header_length()[0]) >>
                         mask.header_length()[1],
        'event_length': (word & mask.event_length()[0]) >>
                        mask.event_length()[1],
        'finish_code': (word & mask.finish_code()[0]) >> mask.finish_code()[1]
    }


def decode_word_two(word, mask):
    """
    Decodes the
       * high component of the event time
       * CFD time
       * CFD Trigger source bit
       * CFD forced trigger bit
    :param word: The word that we're going to decode.
    :param mask: The mask we'll use to decode that word.
    :return: A dictionary containing the decoded information
    """
    return {
        'event_time_high': (word & mask.event_time_high()[0]) >> mask.event_time_high()[1],
        'cfd_fractional_time': (word & mask.cfd_fractional_time()[0]) >>
                               mask.cfd_fractional_time()[1],
        'cfd_trigger_source_bit': (word & mask.cfd_trigger_source()[0]) >>
                                  mask.cfd_trigger_source()[1],
        'cfd_forced_trigger_bit': (word & mask.cfd_forced_trigger()[0]) >>
                                  mask.cfd_forced_trigger()[1]
    }


def decode_word_three(word, mask):
    """
    The final word of the standard header. Contains
        * trapezoidal energy
        * trace length
        * trace out of range flag
    :param word: The word that we're going to decode.
    :param mask: The mask we'll use to decode that word.
    :return: A dictionary containing the decoded information
    """
    return {
        'energy': (word & mask.energy()[0]) >> mask.energy()[1],
        'trace_length': (word & mask.trace_length()[0]) >> mask.trace_length()[1],
        'trace_out_of_range': (word & mask.trace_out_of_range()[0]) >> mask.trace_out_of_range()[1]
    }


def decode_external_timestamps():
    print("decoding timestamps")


def decode_qdc():
    print("decoding qdc")


def decode_energy_sums():
    print("decoding energysums")


def decode_trace():
    print("decoding trace")


class ListModeDataDecoder(threading.Thread):
    """
    Class that loops through a data stream to locate and process Pixie16
    list mode data.
    """

    def __init__(self, stream, mask):
        """
        Constructor
        :param stream: The stream that we'll read data from
        :param mask: the data mask that we'll use to decode data
        """
        threading.Thread.__init__(self)
        self.stream = stream
        self.mask = mask
        self.finished = False

    def run(self):
        """ Decodes data from Pixie16 binary data stream """
        # TODO : Will need to add in decoding of optional header information
        decoded_data_list = []
        for chunk in iter(partial(self.stream.read, data.WORD), b''):
            decoded_data = decode_word_zero(struct.unpack('I', chunk)[0], self.mask)
            decoded_data.update({
                'event_time_low': struct.unpack('I', self.stream.read(data.WORD))[0],
            })
            decoded_data.update(
                decode_word_two(struct.unpack('I', self.stream.read(data.WORD))[0], self.mask))
            decoded_data.update(
                decode_word_three(struct.unpack('I', self.stream.read(data.WORD))[0], self.mask))
            decoded_data_list.append(decoded_data)
        return decoded_data_list



"""
file: list_mode_data_mask_tests.py
brief: Unittests for the list_mode_data_mask class.
author: S. V. Paulauskas
date: April 25, 2019
"""
import unittest
from dolosse.hardware.xia.pixie16.list_mode_data_mask import ListModeDataMask

firmwares = [17562, 20466, 27361, 29432, 30474, 30980, 30981, 34688,
             34703, 35921]

frequencies = [100, 250, 500]


class ListModeDataMaskTestCase(unittest.TestCase):
    """ Class definition for the unittests """

    def test_construction(self):
        """ Tests that the constructor throws errors when we provide bad data"""
        for f in firmwares:
            with self.assertRaises(ValueError):
                ListModeDataMask(1234, f)
        for f in frequencies:
            with self.assertRaises(ValueError):
                ListModeDataMask(f, 10)

    def test_trace_length(self):
        """ Tests that we get the correct trace length for all firmwares"""
        for firmware in firmwares:
            for freq in frequencies:
                if firmware < 34688:
                    self.assertEqual(ListModeDataMask(freq, firmware).trace_length(),
                                     (0xFFFF0000, 16))


if __name__ == '__main__':
    unittest.main()



"""
file: header.py
brief: Defines how to handle an LDF Header buffer
author: S. V. Paulauskas
date: February 04, 2019
"""
import struct

import constants.data as data

class LdfHeader:
    """
    Defines a structure that reads the header information from a ORNL HEAD buffer
    """

    def __init__(self):
        self.run_number = self.max_spill_size = self.run_time = 0
        self.format = self.facility = self.start_date = self.end_date = self.title = ''

    def read_header(self, stream):
        """
        Reads the header out of the provided stream
        :param stream: the stream that contains the header to read
        :return: a dictionary containing the header information
        """
        return {
            'buffer_size': struct.unpack('I', stream.read(data.WORD))[0],
            'facility': stream.read(data.WORD * 2).decode(),
            'format': stream.read(data.WORD * 2).decode(),
            'type': stream.read(data.WORD * 4).decode(),
            'date': stream.read(data.WORD * 4).decode(),
            'title': stream.read(data.WORD * 20).decode(),
            'run_number': struct.unpack('I', stream.read(data.WORD))[0],
        }



"""
file: pixie_net_consumer.py
brief: A Kafka consumer that processes Pixie16 data. Creates a thread for each partition plus extra.
author: S. V. Paulauskas
date: March 30, 2019
"""
from confluent_kafka import Consumer, KafkaError
from statistics import mean
import time


class ServiceExit(Exception):
    """
    Custom exception which is used to trigger the clean exit
    of all running threads and the main program.
    """
    pass


def service_shutdown(signum, frame):
    print('Caught shutdown signal, SIGNUM=%d' % signum)
    raise ServiceExit


consumer = Consumer({'bootstrap.servers': '192.168.1.25:9092',
                     'group.id': 'pixie_net_consumer',
                     'default.topic.config': {
                         'auto.offset.reset': 'latest'},
                     'auto.commit.interval.ms': 500})
consumer.subscribe(['pixie_net'])

timeouts = records = 0
messages__in_interval = errors_in_interval = idle_time_in_interval = 0
stats_interval_start_time = time.time()
message_processing_times = []

TIMEOUT = 0.01
STATS_INTERVAL = 2
try:
    while True:
        msg = consumer.poll(timeout=TIMEOUT)
        if msg is None:
            idle_time_in_interval += TIMEOUT
        elif not msg.error():
            message_processing_start_time = time.time()
            print("Begin processing data buffer")
            print(msg.value())

            message_processing_times.append(time.time() - message_processing_start_time)
        elif msg.error().code() != KafkaError._PARTITION_EOF:
            print(msg.error())

        elapsed_time = time.time() - stats_interval_start_time
        if elapsed_time >= STATS_INTERVAL:
            msg = {'idle_time': idle_time_in_interval}
            if message_processing_times:
                msg['min_message_processsing_time'] = min(message_processing_times)
                msg['ave_message_processing_time'] = mean(message_processing_times)
                msg['max_message_processing_time'] = max(message_processing_times)
            print(msg)
            stats_interval_start_time = time.time()

            idle_time_in_interval = 0
            message_processing_times.clear()
except ServiceExit:
    consumer.close()
    print("Received shutdown call!")
print("See you space cowboy!")



"""
file: list_mode_data_mask.py
brief: Defines data masks for all frequencies, bit resolutions, and revisions
author: S. V. Paulauskas
date: January 16, 2019
"""


class ListModeDataMask:
    """ Defines data masks used to decode data from XIA's Pixie-16 product line. """

    def __init__(self, frequency, firmware):
        if frequency not in [100, 250, 500]:
            raise ValueError(self.__class__.__name__ + ": Please specify a valid frequency "
                                                       "[100, 250, 500].")
        if firmware < 17562:
            raise ValueError(self.__class__.__name__ + ": Please specify a valid firmware revision"
                                                       " greater than 17562.")
        self.frequency = frequency
        self.firmware = firmware

    @staticmethod
    def basic_header_words():
        """ The number of words that make up a basic header with no additional information """
        return 4

    @staticmethod
    def number_of_energy_sum_words():
        """ The number of words that make up the energy sums when they're in the header """
        return 4

    @staticmethod
    def number_of_external_timestamp_words():
        """ The number of words that make up an external timestamp when it's in the header """
        return 2

    @staticmethod
    def number_of_qdc_words():
        """ The number of words that make up the QDC when it's in the header """
        return 8

    @staticmethod
    def channel():
        """ Provides the mask needed to decode the Channel from Word 0 """
        return 0x0000000F, 0

    @staticmethod
    def slot():
        """ Provides the mask needed to decode the Slot from Word 0 """
        return 0x000000F0, 4

    @staticmethod
    def crate():
        """ Provides the mask needed to decode the Crate from Word 0 """
        return 0x00000F00, 8

    @staticmethod
    def header_length():
        """ Provides the mask needed to decode the header_length from Word 0 """
        return 0x0001F000, 12

    @staticmethod
    def finish_code():
        """ Provides the mask needed to decode the Finish Code from Word 0 """
        return 0x80000000, 31

    @staticmethod
    def event_time_high():
        """ Provides the mask needed to decode the Event Time High from Word 2 """
        return 0x0000FFFF, 0

    @staticmethod
    def trace_element():
        """ Pixie16 stores trace information in packed in two elements per word. This provides the
            mask needed to decode the trace element stored in the upper bits of the pair """
        return 0x0000FFFF, 16

    def event_length(self):
        """ Provides the mask needed to decode the Event Length in Word 0 """
        if self.firmware < 29432:
            return 0x3FFE0000, 17
        return 0x7FFE0000, 17

    def cfd_fractional_time(self):
        """
        The CFD Fractional Time always starts on Bit 16 of Header Word 2.
        :return: Tuple of mask and bit for shifting.
        """
        if self.frequency == 100:
            if self.firmware <= 29432:
                return 0xFFFF0000, 16
            return 0x7FFF0000, 16

        if self.frequency == 250:
            if self.firmware == 20466:
                return 0xFFFF0000, 16
            if 27361 <= self.firmware <= 29432:
                return 0x7FFF0000, 16

        if self.frequency == 500:
            if self.firmware >= 29432:
                return 0x1FFF0000, 16

        return 0x3FFF0000, 16  # We'll default to 250, > 29432

    def cfd_size(self):
        """ Provides the CFD Size, which we use to reconstruct the trigger's arrival time """
        if self.frequency == 500:
            return 8192
        if self.frequency == 100:
            if 17562 <= self.firmware < 30474:
                return 65536
            if self.firmware >= 30474:
                return 32768
        if self.frequency == 250:
            if self.firmware == 20466:
                return 65536
            if 27361 <= self.firmware < 30980:
                return 32768
        return 16384

    def cfd_trigger_source(self):
        """ Provides the mask needed to decode the CFD Trigger Source Bit in Word 2 """
        if self.frequency == 250:
            if 27361 <= self.firmware < 30474:
                return 0x80000000, 31
            if self.firmware >= 30474:
                return 0x40000000, 30
        if self.frequency == 500:
            if self.firmware >= 29432:
                return 0xE0000000, 29
        return 0x0, 0

    def cfd_forced_trigger(self):
        """ Provides the mask needed to decode the CDF Forced Trigger Bit in Word 2 """
        if self.frequency in [100, 250]:
            if self.firmware >= 30474:
                return 0x80000000, 31
        return 0x0, 0

    def energy(self):
        """ Provides the mask needed to decode the Energy from Word 3 """
        if self.firmware in [29432, 30474, 30980, 30981]:
            return 0x00007FFF, 0
        return 0x0000FFFF, 0

    def trace_length(self):
        """ Provides the mask needed to decode the Trace Length from Word 3"""
        if 17562 <= self.firmware < 34688:
            return 0xFFFF0000, 16
        return 0x7FFF0000, 16

    def trace_out_of_range(self):
        """ Provides the mask needed to decode the Trace Out of Range Bit """
        if 17562 <= self.firmware < 29432:
            return 0x40000000, 30
        if 29432 <= self.firmware < 34688:
            return 0x00008000, 15
        return 0x80000000, 31

    def generate_value_error_message(self, name):
        """ Generates the string used for ValueErrors """
        return "%s::%s - Could not determine the appropriate mask for firmware (%s) and " \
               "frequency (%s)" % self.__class__.__name__, name, self.firmware, self.frequency



"""
file: list_mode_data_decoder_tests.py
brief: Unittests for the list_mode_data_decoder class.
author: S. V. Paulauskas
date: March 5, 2019
"""
from io import BytesIO
import unittest

import dolosse.hardware.xia.pixie16.list_mode_data_decoder as decoder
import dolosse.hardware.xia.pixie16.list_mode_data_mask as lmdm


# TODO : We'll need to add tests to cover the major firmware revisions.
class ListModeDataDecoderTestCase(unittest.TestCase):
    def setUp(self):
        self.mask = lmdm.ListModeDataMask(250, 30474)

    def test_decode_word_zero(self):
        self.assertEqual({
            'channel': 13,
            'slot': 2,
            'crate': 0,
            'header_length': 4,
            'event_length': 4,
            'finish_code': 0
        }, decoder.decode_word_zero(540717, self.mask))

    def test_decode_word_two(self):
        self.assertEqual({
            'event_time_high': 26001,
            'cfd_fractional_time': 0,
            'cfd_trigger_source_bit': 0,
            'cfd_forced_trigger_bit': 0
        }, decoder.decode_word_two(26001, self.mask))

    def test_decode_word_three(self):
        self.assertEqual({
            'energy': 2345,
            'trace_length': 0,
            'trace_out_of_range': 0
        }, decoder.decode_word_three(2345, self.mask))

    def test_run(self):
        self.assertEqual(
            decoder.ListModeDataDecoder(
                BytesIO(b'\x2D\x40\x08\x00\x15\xCD\x5B\x07\x91\x65\x00\x00\x29\x09\x00\x00'),
                self.mask).run(),
            [{
                'channel': 13,
                'slot': 2,
                'crate': 0,
                'header_length': 4,
                'event_length': 4,
                'finish_code': 0,
                'event_time_low': 123456789,
                'event_time_high': 26001,
                'cfd_fractional_time': 0,
                'cfd_trigger_source_bit': 0,
                'cfd_forced_trigger_bit': 0,
                'energy': 2345,
                'trace_length': 0,
                'trace_out_of_range': 0
            }])


if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dolosse",
    version="0.0.0",
    author="S. V. Paulauskas and the Dolosse Collaboration",
    author_email="stan@projectscience.tech",
    description="A Kafka based data acquisition and analysis framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dolosse/dolosse",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Physics",
        "Operating System :: OS Independent",
    ],
)


"""
file: data.py
brief: A namespace containing constants related to data management
author: S. V. Paulauskas
date: January 27, 2019
"""

# There are 4 bytes in 1 32-bit word, this will be the basis for our reads.
WORD = 4

# Here we define some of the keywords that define different types of blocks in the binary data.
DIR_BLOCK = b'DIR '
HEAD_BLOCK = b'HEAD'
DATA_BLOCK = b'DATA'
PAC_BLOCK = b'PAC '
SCALAR_BLOCK = b'SCAL'
DEADTIME_BLOCK = b'DEAD'
END_OF_FILE = b'EOF '
END_OF_BUFFER = b'EOB '
BUFFER_PADDING = b'\xff\xff\xff\xff'


"""
file: constants.py
brief: Constants for the LDF class.
author: S. V. Paulauskas
date: February 05, 2019
"""
LDF_DATA_BUFFER_SIZE = 8193
LDF_DATA_BUFFER_SIZE_WITHOUT_HEADER = 8192
MINIMUM_CHUNK_SIZE = 20